# (C) Copyright SuperDARN Canada, University of Saskatchewan
# Author: Marina Schmidt
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.
#
# Modifications:
#

# Standard accepted IUGG global average of Earth's radius (km)
Re = 6371
# Earth radius at the equator (km)
EARTH_EQUATORIAL_RADIUS = 6378.137

# Speed of light (m/s)
C = 2.99792458e+8
